"""
Using ContextInstanceMixin from aiogram see:
https://github.com/aiogram/aiogram/blob/dev-2.x/aiogram/utils/mixins.py
"""

import contextvars
from typing import TypeVar, Type

T = TypeVar('T')


class ContextInstanceMixin:
    def __init_subclass__(cls, **kwargs):
        cls.__context_instance = contextvars.ContextVar('instance_' + cls.__name__)
        return cls

    @classmethod
    def get_current(cls: Type[T], no_error=True) -> T:
        if no_error:
            return cls.__context_instance.get(None)
        return cls.__context_instance.get()

    @classmethod
    def set_current(cls: Type[T], value: T):
        if not isinstance(value, cls):
            raise TypeError(f"Value should be instance of '{cls.__name__}' not '{type(value).__name__}'")
        cls.__context_instance.set(value)



"""
Based on: https://github.com/MrMrRobat/AnyStrEnum
"""

import sys
from enum import Enum, EnumMeta, _EnumDict, auto
from types import FunctionType
from typing import List, Callable, AnyStr, Set, TypeVar, Type, Any

SEP_ATTR = "__sep__"
CONVERTER_ATTR = "__converter__"

TYPE_BLACKLIST = FunctionType, property, type, classmethod, staticmethod


class StrItem:
    __slots__ = 'sep', 'converter'

    def __init__(self, sep: AnyStr = None, converter: Callable[[str], str] = None):
        self.sep = sep
        self.converter = converter

    def generate_value(self, name: str) -> str:
        if self.converter:
            name = self.converter(name)
        if self.sep:
            name = name.replace('_', self.sep)

        return name


class BaseStrEnum(str, Enum):
    __sep__: str = None
    __converter__: Callable[[str], str] = None
    __return_missing__: bool = False

    def _generate_next_value_(*_):
        return auto()

    @classmethod
    def filter(cls,
               contains: AnyStr = None, *,
               contained_in: AnyStr = None,
               startswith: AnyStr = None,
               endswith: AnyStr = None,
               case_sensitive: bool = False,
               intersection: bool = True,
               inverse: bool = False) -> Set['StrEnum']:
        """
        :param contains: filter all enum members which are contain some substring
        :param startswith: filter all enum members which are start with some substring
        :param endswith: filter all enum members which are end with some substring
        :param contained_in: filter all enum members which are substrings of some string
        :param case_sensitive: defines whether found values must match case of given string
        :param inverse: if True, all enum members except found will be returned
        :param intersection: indicates whether function should return all found objects or their interception
        :return: all found enums
        """

        def prepare(value):
            if case_sensitive:
                return value
            return value.lower()

        found_sets: List[set] = []
        if contains:
            contains = prepare(contains)
            found_sets.append({e for e in cls if contains in prepare(e)})
        if startswith:
            startswith = prepare(startswith)
            found_sets.append({e for e in cls if prepare(e).startswith(startswith)})
        if endswith:
            endswith = prepare(endswith)
            found_sets.append({e for e in cls if prepare(e).endswith(endswith)})
        if contained_in:
            contained_in = prepare(contained_in)
            found_sets.append({e for e in cls if prepare(e) in contained_in})

        if not found_sets:
            return set()

        if intersection:
            found = found_sets[0].intersection(*found_sets[1:])
        else:
            found = found_sets[0].union(*found_sets[1:])

        if inverse:
            return {e for e in cls} - found

        return found

    def __str__(self):
        return self.value


class StrEnumMeta(EnumMeta):
    # It's here to avoid 'got an unexpected keyword argument' TypeError
    @classmethod
    def __prepare__(mcs, *args, sep: AnyStr = None, converter: Callable[[str], str] = None, **kwargs):
        return super().__prepare__(*args, **kwargs)

    def __new__(mcs, cls, bases, class_dict, sep: AnyStr = None, converter: Callable[[str], str] = None):
        # In Python 3.8 the signature of 'EnumMeta._get_mixins_' was changed from (bases) to (class_name, bases)
        if sys.version_info >= (3, 8):  # (class_name, bases)
            mixin_type, base_enum = mcs._get_mixins_(cls, bases)
        else:  # Fallback to (bases) signature
            mixin_type, base_enum = mcs._get_mixins_(bases)
        if not issubclass(base_enum, BaseStrEnum):
            raise TypeError(f'Unexpected Enum type \'{base_enum.__name__}\'. '
                            f'Only {BaseStrEnum.__name__} and its subclasses are allowed')
        if not issubclass(mixin_type, (str, bytes)):
            raise TypeError(f'Unexpected mixin type \'{mixin_type.__name__}\'. '
                            f'Only str, bytes and their subclasses are allowed')

        # Trying to get sep and converter from class dict and base enum class
        if sep is None:
            sep = class_dict.get(SEP_ATTR) or base_enum.__sep__
        if converter is None:
            converter = class_dict.get(CONVERTER_ATTR) or base_enum.__converter__

        item = StrItem(sep=sep, converter=converter)
        new_class_dict = _EnumDict()
        for name, type_hint in class_dict.get('__annotations__', {}).items():
            if name.startswith('_') or name in class_dict:
                continue
            mcs.check_type_equals(type_hint, mixin_type)
            value = item.generate_value(name)
            new_class_dict[name] = value
            mcs.check_type_equals(type(value), mixin_type)

        for name, value in class_dict.items():
            if isinstance(value, StrItem):
                value = value.generate_value(name)
            if not (name.startswith('_') or isinstance(value, TYPE_BLACKLIST)):
                mcs.check_type_equals(type(value), mixin_type)

            new_class_dict[name] = value

        new_class_dict[SEP_ATTR] = sep
        new_class_dict[CONVERTER_ATTR] = converter

        return super().__new__(mcs, cls, bases, new_class_dict)

    def __call__(cls, value, **kwargs):
        """
        For ability to check values by 'isinstance'
        (only if their type same as enums base type)

        >>> class MyEnum(StrEnum):
        ...     __return_missing__ = True
        ...     member1: str
        ...     member2: str
        ...
        >>> value_to_check = 'invalid_member'
        >>> if not isinstance(MyEnum(value_to_check), MyEnum):
        ...     print(f'{value_to_check} is not a valid MyEnum')
        invalid_member is not a valid MyEnum

        :param value: value to check
        :param kwargs: named args for EnumMeta
        """

        try:
            return super().__call__(value, **kwargs)
        except ValueError:
            if not kwargs and cls.__return_missing__ and type(value) is cls._member_type_:
                return value
            raise

    @staticmethod
    def check_type_equals(type_to_check: Any, allowed_type: Type[Any]):
        if type_to_check is not allowed_type:
            raise TypeError(f'Unexpected type {getattr(type_to_check, "__name__", type_to_check)}'
                            f', allowed type: {allowed_type.__name__}')


class StrEnum(BaseStrEnum, metaclass=StrEnumMeta):
    __return_missing__ = True



text = '''
InvalidS3BucketException
The provided Amazon S3 bucket name is invalid.

HTTP Status Code: 400

InvalidS3KeyException
The provided Amazon S3 key prefix is invalid. Please provide a valid S3 object key name.

HTTP Status Code: 400

InvalidSampleRateException
The specified sample rate is not valid.

HTTP Status Code: 400

InvalidSnsTopicArnException
The provided SNS topic ARN is invalid. Please provide a valid SNS topic ARN and try again.

HTTP Status Code: 400

InvalidSsmlException
The SSML you provided is invalid.

HTTP Status Code: 400

LanguageNotSupportedException
The language specified is not currently supported by Amazon Polly in this capacity.

HTTP Status Code: 400

LexiconNotFoundException
Amazon Polly can't find the specified lexicon.

HTTP Status Code: 404

MarksNotSupportedForFormatException
Speech marks are not supported for the OutputFormat selected.

HTTP Status Code: 400

ServiceFailureException
An unknown condition has caused a service failure.

HTTP Status Code: 500

SsmlMarksNotSupportedForTextTypeException
SSML speech marks are not supported for plain text-type input.

HTTP Status Code: 400

TextLengthExceededException
The value of the "Text" parameter is longer than the accepted limits. 

HTTP Status Code: 400\
'''
t = text.replace('\'', '\\\'').split('\n\n')
for n, s in enumerate(t):
    try:
        exception, message = s.strip().split('\n')
        print(f"class {exception}(BasePollyException):\n    msg: str = '{message}'")
    except ValueError:
        print('    http_code: int = ' + s[-3:])
        print('\n')



from typing import Union, List

from ..types import LanguageCode, Alphabet

LEXICON_TEMPLATE = '''\
<?xml version="1.0" encoding="UTF-8"?>
<lexicon version="1.0" 
      xmlns="http://www.w3.org/2005/01/pronunciation-lexicon"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
      xsi:schemaLocation="http://www.w3.org/2005/01/pronunciation-lexicon 
        http://www.w3.org/TR/2007/CR-pronunciation-lexicon-20071212/pls.xsd"
      alphabet="{alphabet}" 
      xml:lang="{lang}">
  {lexemes}
</lexicon>
'''

LEXEME_TEMPLATE = '<lexeme>%s\n  </lexeme>'
GRAPHEME_TEMPLATE = '\n     <grapheme>%s</grapheme>'
PHONEME_TEMPLATE = '\n     <phoneme>%s</phoneme>'
ALIAS_TEMPLATE = '\n     <alias>%s</alias>'


def new_lexeme(grapheme: str, *, phoneme: str = None, alias: str = None):
    result = GRAPHEME_TEMPLATE % grapheme

    if phoneme is not None:
        result += PHONEME_TEMPLATE % phoneme
    if alias is not None:
        result += ALIAS_TEMPLATE % alias

    return LEXEME_TEMPLATE % result


def new_lexicon(alphabet: Alphabet, lang: LanguageCode,
                lexemes: Union[str, dict, List[Union[dict, str]]]):
    lexemes_xml = lexemes_to_xml(lexemes)

    if isinstance(alphabet, Alphabet):
        alphabet = alphabet.value
    if isinstance(lang, LanguageCode):
        lang = lang.value

    return LEXICON_TEMPLATE.format(alphabet=alphabet, lang=lang, lexemes=lexemes_xml)


def lexemes_to_xml(lexemes):
    if isinstance(lexemes, list):
        lexemes_xml = ''.join(lexemes_to_xml(l) for l in lexemes)
    elif isinstance(lexemes, dict):
        lexemes_xml = new_lexeme(**lexemes)
    elif hasattr(lexemes, 'to_str'):
        lexemes_xml = lexemes.to_str()
    elif isinstance(lexemes, str):
        lexemes_xml = lexemes
    else:
        raise ValueError(f'Unsupported lexemes type {type(lexemes)}')

    return lexemes_xml



from .actions import (
    breath, clean_text_from_ssml_tags, drc, emphasis,
    lang, mark, paragraph, pause, phoneme, prosody,
    say_as, sentence, soft, ssml_text, sub, timbre, whisper
)
from .params import Level, Volume, Pitch, Interpretation, Duration, DateFormat, Rate, Frequency, Strength, Alphabet

__all__ = [
    'breath',
    'clean_text_from_ssml_tags',
    'drc',
    'emphasis',
    'lang',
    'mark',
    'paragraph',
    'pause',
    'phoneme',
    'prosody',
    'say_as',
    'sentence',
    'soft',
    'ssml_text',
    'sub',
    'timbre',
    'whisper',
    'Alphabet',
    'DateFormat',
    'Duration',
    'Frequency',
    'Interpretation',
    'Pitch',
    'Level',
    'Volume',
    'Rate',
    'Strength'
]



import datetime

from . import types
from . import utils
from .polly.polly import Polly
from .utils import exceptions

__all__ = [
    'Polly',
    'exceptions',
    'types',
    'utils'
]

__version__ = '0.2.2'
__api_date__ = datetime.date(2019, 7, 30)  # July 30, 2019



import re
from typing import ClassVar, Union

from aiohttp import ClientResponse

exception_pattern = re.compile(r'([A-Z][a-z0-9]+){2,}')

PolyExceptionType = ClassVar['PollyAPIException']


class BasePollyException(Exception):
    msg = 'Exception occurred'
    match = None

    def __init__(self,
                 message: str = None,
                 url: str = None,
                 payload: dict = None,
                 response: ClientResponse = None,
                 content: Union[dict, str, bytes, ClientResponse] = None,
                 cause: ClassVar[Exception] = None):

        if not message:
            message = self.msg
        if cause is not None:
            message += f'\n\nCaused by: {cause.__class__.__name__}: {cause}'
        if url is not None:
            message += f'\n\nRequest url: {url}'
        if payload is not None:
            message += f'\n\nPayload: {payload}'
        if response is not None:
            message += f'\n\nResponse: {response}'
        if content is not None:
            message += f'\n\nContent: {content}'

        super().__init__(message)
        self.url = url
        self.payload = payload
        self.response = response
        self.cause = cause


class AioHTTPException(BasePollyException):
    msg = 'aiohttp throws exception:'


class JSONDecodeException(BasePollyException):
    msg = 'Unable to decode JSON'


class ResponseTypeException(BasePollyException):
    msg = 'Server returned response with unexpected content type'


class ResponseValueException(BasePollyException):
    msg = match = 'API send response with unexpected values'


class PollyAPIException(BasePollyException):
    msg = 'API returned exception'


class BadRequestException(PollyAPIException):
    http_code = 400
    retry = False


class AccessDeniedException(PollyAPIException):
    http_code = 403
    retry = False


class NotFoundException(PollyAPIException):
    http_code = 404
    retry = False


class ConflictException(PollyAPIException):
    http_code = 409
    retry = False


class LimitExceededException(PollyAPIException):
    http_code = 429
    retry = False


class TooManyRequestsException(PollyAPIException):
    http_code = 429
    retry = False


class ServiceFailureException(PollyAPIException):
    msg = match = 'An unknown condition has caused a service failure.'
    http_code: int = 500
    retry = True


class ServiceUnavailableException(PollyAPIException):
    http_code = 503
    retry = True


class BadGatewayException(PollyAPIException):
    http_code = 502
    retry = True


class EndpointRequestTimedOutException(PollyAPIException):
    http_code = 504
    retry = True


class InvalidS3BucketException(BadRequestException):
    msg = match = 'The provided Amazon S3 bucket name is invalid.'


class InvalidS3KeyException(BadRequestException):
    msg = match = 'The provided Amazon S3 key prefix is invalid. Please provide a valid S3 object key name.'


class InvalidSampleRateException(BadRequestException):
    msg = match = 'The specified sample rate is not valid.'


class InvalidSnsTopicArnException(BadRequestException):
    msg = match = 'The provided SNS topic ARN is invalid. Please provide a valid SNS topic ARN and try again.'


class InvalidSSMLException(BadRequestException):
    msg = match = 'The SSML you provided is invalid.'


class LanguageNotSupportedException(BadRequestException):
    msg = match = 'The language specified is not currently supported by Amazon Polly in this capacity.'


class LexiconNotFoundException(NotFoundException):
    msg = match = 'Amazon Polly can\'t find the specified lexicon.'


class MarksNotSupportedForFormatException(BadRequestException):
    msg = match = 'Speech marks are not supported for the OutputFormat selected.'


class SSMLMarksNotSupportedForTextTypeException(BadRequestException):
    msg = match = 'SSML speech marks are not supported for plain text-type input.'


class TextLengthExceededException(BadRequestException):
    msg = match = 'Maximum text length has been exceeded'


class InvalidTaskIdException(BadRequestException):
    msg = match = 'The provided Task ID is not valid. Please provide a valid Task ID and try again.'


class SynthesisTaskNotFoundException(BadRequestException):
    msg = match = 'The Speech Synthesis task with requested Task ID cannot be found.'


class InvalidNextTokenException(BadRequestException):
    msg = match = 'The NextToken is invalid. Verify that it\'s spelled correctly, and then try again.'


class InvalidLexiconException(BadRequestException):
    msg = match = 'Amazon Polly can\'t find the specified lexicon.'


class LexiconSizeExceededException(BadRequestException):
    msg = match = 'The maximum size of the specified lexicon would be exceeded by this operation.'


class MaxLexemeLengthExceededException(BadRequestException):
    msg = match = 'The maximum size of the lexeme would be exceeded by this operation.'


class MaxLexiconsNumberExceededException(BadRequestException):
    msg = match = 'The maximum number of lexicons would be exceeded by this operation.'


class UnsupportedPLSAlphabetException(BadRequestException):
    msg = match = 'The alphabet specified by the lexicon is not a supported alphabet. Valid values are x-sampa and ipa.'


class UnsupportedPLSLanguageException(BadRequestException):
    msg = match = 'Unsupported PLS language'


class SignatureDoesNotMatchException(PollyAPIException):
    msg = match = 'The request signature we calculated does not match the signature you provided.'


class ValidationErrorExceptions(PollyAPIException):
    msg = match = 'validation error detected'


class MissingAuthTokenException(PollyAPIException):
    msg = match = 'Missing Authentication Token'


class InvalidPLSLexiconException(BadRequestException):
    msg = match = 'Invalid PLS Lexicon'


class InvalidSSMLRequestException(BadRequestException):
    msg = match = 'Invalid SSML request'


EXCEPTIONS = set()
for k, v in dict(globals()).items():
    try:
        if issubclass(v, BasePollyException):
            EXCEPTIONS.add(v)
    except TypeError:
        pass

BASE_EXCEPTIONS = (
    BadRequestException,
    AccessDeniedException,
    NotFoundException,
    ConflictException,
    LimitExceededException,
    TooManyRequestsException,
    BadGatewayException,
    ServiceUnavailableException,
    EndpointRequestTimedOutException
)


def get_exception(exception_header: str, response_code: int, message: str = None) -> PolyExceptionType:
    """
    Exception header example:
    'UnrecognizedClientException:http://internal.amazon.com/coral/com.amazon.coral.service/'
    'UnauthorizedException'
    """
    # Trying to find and return existing exception
    if message:
        match_exception = next((exc for exc in EXCEPTIONS if exc.match and exc.match in message), None)
        if match_exception:
            return match_exception

    # Trying to resolve AWS exception name from exception header if it exists
    exc_name = None
    if exception_header:
        match = exception_pattern.match(exception_header)
        if match:
            exc_name = match.group()

            # Last try to find existing exception by its name (in case match text has been changed)
            match_exception = next((exc for exc in EXCEPTIONS if exc.__name__.lower() == exc_name.lower()), None)
            if match_exception:
                return match_exception

    # If exception not found, creating new one basing on response_code
    __bases__ = next((exc for exc in BASE_EXCEPTIONS if exc.http_code == response_code), PollyAPIException)

    return type(
        exc_name or __bases__.__name__,
        (__bases__,),
        {'msg': message or __bases__.msg}
    )



"""
See documentation: https://docs.aws.amazon.com/en_us/polly/latest/dg/supported-ssml.html
"""

from ..strenum import StrEnum
from ...types import Alphabet

__all__ = [
    'Alphabet',
    'Frequency',
    'Duration',
    'Level',
    'Strength',
    'Volume',
    'Pitch',
    'Rate',
    'Interpretation',
    'DateFormat'
]


class DashParamEnum(StrEnum):
    __sep__ = '-'


class Level(DashParamEnum):
    """
    :enum Strong: Increases the volume and slows the speaking rate so that the speech is louder and slower.
    :enum Moderate: Increases the volume and slows the speaking rate, but less than strong. Moderate is the default.
    :enum Reduced: Decreases the volume and speeds up the speaking rate. Speech is softer and faster.
    """
    strong: str
    moderate: str
    reduced: str


class Strength(DashParamEnum):
    """
    :enum none: No pause. Use none to remove a normally occurring pause, such as after a period.
    :enum x-weak: Has the same strength as none, no pause.
    :enum weak: Sets a pause of the same duration as the pause after a comma.
    :enum medium: Has the same strength as weak.
    :enum strong: Sets a pause of the same duration as the pause after a sentence.
    :enum x-strong: Sets a pause of the same duration as the pause after a paragraph.
    """
    none: str
    x_weak: str
    weak: str
    medium: str
    strong: str
    x_strong: str


class Volume(DashParamEnum):
    """
    :enum default: Resets volume to the default level for the current voice.
    :enums: silent, x-soft, soft, medium, loud, x-loud: Sets the volume to a predefined value for the current voice.
    """
    default: str
    silent: str
    x_soft: str
    soft: str
    medium: str
    loud: str
    x_loud: str


class Rate(DashParamEnum):
    """
    :enums: x-slow, slow, medium, fast,x-fast. Sets the pitch to a predefined value for the selected voice.
    """
    x_slow: str
    slow: str
    medium: str
    fast: str
    x_fast: str


class Pitch(DashParamEnum):
    """
    :enum default: Resets pitch to the default level for the current voice.
    :enums x-low, low, medium, high, x-high: Sets the pitch to a predefined value for the current voice.
    """
    default: str
    x_low: str
    low: str
    medium: str
    high: str
    x_high: str


class Interpretation(DashParamEnum):
    """
    :enum characters or spell-out: Spells out each letter of the text, as in a-b-c.
    :enum cardinal or number: Interprets the numerical text as a cardinal number, as in 1,234.
    :enum ordinal: Interprets the numerical text as an ordinal number, as in 1,234th.
    :enum digits: Spells out each digit individually, as in 1-2-3-4.
    :enum fraction: Interprets the numerical text as a fraction.
    :enum unit: Interprets a numerical text as a measurement.
    :enum date: Interprets the text as a date. The format of the date must be specified with the format attribute.
    :enum time: Interprets the numerical text as duration, in minutes and seconds, as in 1'21".
    :enum address: Interprets the text as part of a street address.
    :enum expletive: "Beeps out" the content included within the tag.
    :enum telephone: Interprets the numerical text as a 7-digit or 10-digit telephone number, as in 2025551212.
    """
    characters: str
    spell_out: str
    cardinal: str
    number: str
    ordinal: str
    digits: str
    fraction: str
    unit: str
    date: str
    time: str
    address: str
    expletive: str
    telephone: str


class DateFormat(DashParamEnum):
    """
    :enum mdy: Month-day-year.
    :enum dmy: Day-month-year.
    :enum ymd: Year-month-day.
    :enum md: Month-day.
    :enum dm: Day-month.
    :enum ym: Year-month.
    :enum my: Month-year.
    :enum d: Day.
    :enum m: Month.
    :enum y: Year.
    :enum yyyymmdd: Year-month-day.
    """

    mdy: str
    dmy: str
    ymd: str
    md: str
    dm: str
    ym: str
    my: str
    d: str
    m: str
    y: str
    yyyymmdd: str


class Frequency(DashParamEnum):
    """Controls how often breathing sounds occur in the text."""
    default: str
    x_low: str
    low: str
    medium: str
    high: str
    x_high: str


class Duration(DashParamEnum):
    """Controls the length of the breath."""
    default: str
    x_short: str
    short: str
    medium: str
    long: str
    x_long: str


class SpeechPart(StrEnum, converter=lambda name: f'amazon:{name}'):
    """
    :enum amazon:VB: interprets the word as a verb (present simple).
    :enum: amazon:VBD: interprets the word as past tense or as a past participle.
    :enum amazon:SENSE_1: uses the non-default sense of the word when present.
        For example, the noun "bass" is pronounced differently depending on its meaning.
        The default meaning is the lowest part of the musical range.
        The alternate meaning is a species of freshwater fish,
        also called "bass" but pronounced differently.
        Using <w role="amazon:SENSE_1">bass</w> renders the non-default pronunciation
        (freshwater fish) for the audio text.

    """
    VB: str
    VBD: str
    SENSE_1: str



import asyncio
from typing import List, Union

from .base import BasePollyObject
from .params import Gender, VoiceID, LanguageCode, SpeechMarkTypes
from .speech import Speech, SpeechMarksList

__all__ = ['Voice', 'VoicesList']


class Voice(BasePollyObject):
    additional_language_codes: List[str] = None
    gender: Gender
    id: VoiceID
    language_code: LanguageCode
    language_name: str
    name: str

    async def synthesize_speech(self, text: str,
                                output_format: str = None,
                                sample_rate: str = None,
                                speech_mark_types: str = None,
                                text_type: str = None,
                                language_code: str = None,
                                lexicon_names: str = None,
                                auto_convert: bool = None,
                                **convert_params
                                ) -> Union[Speech, SpeechMarksList]:
        return await self.polly.synthesize_speech(text=text,
                                                  voice_id=self.id,
                                                  output_format=output_format,
                                                  sample_rate=sample_rate,
                                                  speech_mark_types=speech_mark_types,
                                                  text_type=text_type,
                                                  language_code=language_code,
                                                  lexicon_names=lexicon_names,
                                                  auto_convert=auto_convert,
                                                  **convert_params)


class VoicesList(BasePollyObject):
    next_token: str = None
    voices: List[Voice]

    def __iter__(self):
        return iter(self.voices)

    def __getitem__(self, item):
        return self.voices[item]

    async def synthesize_speech(self, text: str,
                                output_format: str = None,
                                sample_rate: str = None,
                                speech_mark_types: List[Union[SpeechMarkTypes, str]] = None,
                                text_type: str = None,
                                language_code: str = None,
                                lexicon_names: str = None,
                                auto_convert: bool = None,
                                **convert_params
                                ) -> Union[List[Speech], List[SpeechMarksList]]:
        return await asyncio.gather(
            *(
                voice.synthesize_speech(text=text,
                                        output_format=output_format,
                                        sample_rate=sample_rate,
                                        speech_mark_types=speech_mark_types,
                                        text_type=text_type,
                                        language_code=language_code,
                                        lexicon_names=lexicon_names,
                                        auto_convert=auto_convert,
                                        **convert_params)
                for voice in self.voices
            )
        )



from botocore.credentials import Credentials
from botocore.session import Session


def get_credentials(access_key: str = None, secret_key: str = None, ):
    if access_key and secret_key:
        return Credentials(access_key, secret_key)

    return Session().get_credentials()



import asyncio
import logging
import ssl
from http import HTTPStatus
from typing import Union, Tuple, Optional

import aiohttp
import certifi
from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest

from . import config
from ..types import ContentType
from ..types.method import Method
from ..utils import json, case
from ..utils.credentials import get_credentials
from ..utils.exceptions import get_exception, ResponseTypeException, JSONDecodeException, AioHTTPException
from ..utils.mixins import ContextInstanceMixin

log = logging.getLogger('aiopolly')


class AmazonAPIClient(ContextInstanceMixin):
    _service_name = config.SERVICE_NAME
    _base_url_template = config.BASE_URL_TEMPLATE

    _exception_header = config.EXCEPTION_HEADER

    _trust_api_responses = config.TRUST_API_RESPONSES
    _convert_to_snake = config.CONVERT_TO_SNAKE_CASE

    def __init__(self, region: str,
                 access_key: Optional[str],
                 secret_key: Optional[str],
                 loop: Optional[asyncio.AbstractEventLoop]):

        self.region = region

        self.base_url = self._base_url_template.format(
            service_name=self._service_name,
            region=self.region,
        )

        self.__signer = SigV4Auth(
            credentials=get_credentials(access_key, secret_key),
            service_name=self._service_name,
            region_name=self.region
        )

        if not loop:
            loop = asyncio.get_event_loop()
        self.loop = loop

        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = aiohttp.TCPConnector(ssl=ssl_context, loop=self.loop)
        self.session = aiohttp.ClientSession(connector=connector, loop=self.loop, json_serialize=json.dumps)

        self.set_current(self)

    def __get_signed_headers(self, url: str, request_method: str = None, payload: str = None):
        request = AWSRequest(method=request_method, url=url, data=payload)
        self.__signer.add_auth(request)

        return dict(request.headers.items())

    async def request(self, method: Method, payload: dict = None, params: dict = None
                      ) -> Tuple[Union[dict, bytes, None], aiohttp.ClientResponse]:

        log.debug('Preparing request to API. method: %s, paylod: %s, params: %s', method, payload, params)

        url = method.get_url(self.base_url, params)
        request_method = method.request_method

        if payload is not None:
            payload_json = json.dumps(payload)
        else:
            payload_json = None

        headers = self.__get_signed_headers(url, request_method, payload_json)

        try:
            log.debug('Sending request to API "%s": [%s]', url, payload_json)
            async with self.session.request(method=request_method,
                                            url=url,
                                            data=payload_json,
                                            headers=headers) as response:
                content = await self.get_content(url, method, payload, response)
        except aiohttp.ClientError as e:
            raise AioHTTPException(url=url, payload=payload, cause=e)

        return content, response

    async def get_content(self, url: str, method: Method, payload: dict, response: aiohttp.ClientResponse):
        log.debug('Getting content for "%s": [%d]', url, response.status)

        if response.content_type == ContentType.application_json:
            try:
                content = await response.json()
            except ValueError as e:
                raise JSONDecodeException(url=url, payload=payload, response=response, cause=e)
        elif response.content_type == ContentType.application_x_json_stream:
            content = await response.content.read()
        else:
            content = await response.read()

        log.debug('Content for "%s": [%d] "%r"', url, response.status, content)

        if response.status not in range(HTTPStatus.OK, HTTPStatus.BAD_REQUEST):
            await self.raise_api_exception(url, payload, response, content)
        if method.no_data_on_success:
            return content
        elif response.content_type not in method.expected_content_types:
            raise ResponseTypeException(url=url, payload=payload, response=response, content=content)

        return content

    async def raise_api_exception(self, url: str, payload: Union[dict, None], response: aiohttp.ClientResponse,
                                  content):
        if response.content_type == ContentType.application_json:
            error_message = content.get('message')
        elif content:
            error_message = content
        else:
            try:
                error_message = await response.text()
            except ValueError:
                error_message = await response.read()

        api_exception = get_exception(
            response.headers.get(self._exception_header),
            response.status,
            error_message
        )
        raise api_exception(url=url, payload=payload, content=content, response=response)

    async def close(self):
        await self.session.close()



import logging
import re
from asyncio import create_subprocess_shell, subprocess
from typing import Sequence

from .base import BaseConverter
from ...types import Speech

__all__ = ['OpusConverter']

INPUT_PARAMS = {
    'pcm': '-f s16le -ar 16000 -ac 1 ',
}

FILTERS = {
    'atempo': 'atempo',
    'volume': 'volume',
}

DEFAULT_BITRATE = {
    'mp3': 48,
    'ogg_vorbis': 35,
    'pcm': 128
}

DEFAULT_SAMPLE_RATE = {
    'mp3': 24000,
    'ogg_vorbis': 24000,
    'pcm': 16000
}

COMMAND_TEMPLATE = '''\
ffmpeg {in_params}-i pipe:0 {filters} -ar {out_sample_rate} -f opus -acodec libopus -b:a {out_bitrate}k pipe:1
'''


def check_ffmpeg():
    import subprocess
    try:
        proc = subprocess.Popen(
            ['ffmpeg', '-encoders'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout, stderr = proc.communicate()
        encoders = stdout.decode()
    except Exception as e:
        raise RuntimeError(f'Unable to access ffmpeg on your system: {e}')

    if 'libopus' not in encoders:
        raise RuntimeError(f'Unable to find libopus in ffmpeg encoders:\n{stderr.decode()}')


class OpusConverter(BaseConverter):
    """
    This is sample class which provide ability to convert your speech to ogg_opus
    You can create your own basing on BaseConverter and use it.

    To use this converter you need FFMpeg with libopus installed on your system

    Default encode params:
        :param out_bitrate: preferred out_bitrate
        :param speed:
        :param tempo:
        :param volume:
    """

    def __init__(self,
                 out_bitrate: int = None,
                 sample_rate: int = None,
                 auto_convert: bool = True,
                 keep_original: bool = False,
                 excluded_filters: Sequence[str] = (),
                 **default_filters):

        check_ffmpeg()

        self.auto_convert = auto_convert
        self.keep_original = keep_original

        self.defaults = dict(out_bitrate=out_bitrate, sample_rate=sample_rate)
        self.default_filters = default_filters
        self.excluded_filters = excluded_filters

    async def convert(self, speech: Speech, out_bitrate: str = None, sample_rate: int = None, **filters):
        out_bitrate = out_bitrate or self.defaults.get('out_bitrate')
        sample_rate = sample_rate or self.defaults.get('sample_rate')

        cmd_params = self._get_cmd_params(speech, out_bitrate, sample_rate, filters)
        cmd = COMMAND_TEMPLATE.format(**cmd_params)

        bytestream, info = await self._execute(cmd, speech.audio_stream)
        duration_in_seconds = self._find_duration(str(info))

        if not self.keep_original:
            speech.audio_stream = bytestream
        speech.converted_stream = bytestream

        speech.converted = True
        speech.converted_params = {'to_format': 'ogg_opus', 'duration_in_seconds': duration_in_seconds, **cmd_params}

        return speech

    def _get_cmd_params(self, speech: Speech, out_bitrate: int, sample_rate: int, filters):
        filters = ' '.join(f'{key}={value}'
                           for key, value in filters.items()
                           if key not in self.excluded_filters)
        return {
            'out_bitrate': out_bitrate or DEFAULT_BITRATE.get(speech.output_format),
            'out_sample_rate': sample_rate or DEFAULT_SAMPLE_RATE.get(speech.output_format),
            'in_params': INPUT_PARAMS.get(speech.output_format, ''),
            'filters': filters and '-filter:a ' + filters
        }

    @staticmethod
    async def _execute(cmd: str, bytestream: bytes):
        proc = await create_subprocess_shell(
            cmd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        stdout, stderr = await proc.communicate(bytestream)
        logging.debug(f'[{cmd!r} exited with {proc.returncode}]')

        if stderr and not stdout:
            raise RuntimeError(f'[{cmd!r} exited with {proc.returncode}]\n{stderr.decode()}')
        elif not stderr:
            raise RuntimeError(f'[{cmd!r} exited with {proc.returncode}]\n')

        logging.debug(f'[stderr]\n{stderr.decode()}')

        return stdout, stderr.decode()

    @staticmethod
    def _find_duration(s):
        try:
            time = re.findall(r'\d{1,3}:\d{1,3}:\d{1,3}', s)[0]
            hours, minutes, seconds = [i for i in time.split(':')]
            duration_in_seconds = int(hours) * 60 ** 2 + int(minutes) * 60 + round(float(seconds))
        except (ValueError, IndexError):
            duration_in_seconds = None

        return duration_in_seconds



import datetime
from typing import List

from .base import BasePollyObject
from .params import AudioFormat

__all__ = ['SynthesisTask', 'SynthesisTasksList']


class SynthesisTask(BasePollyObject):
    creation_time: datetime.datetime
    language_code: str
    lexicon_names: List[str]
    output_format: AudioFormat
    output_uri: str
    request_characters: int
    sample_rate: str
    sns_topic_arn: str
    speech_mark_types: List[str]
    task_id: str
    task_status: str
    task_status_reason: str
    text_type: str
    voice_id: str


class SynthesisTasksList(BasePollyObject):
    synthesis_tasks: List[SynthesisTask]
    next_token: str = None

    def __iter__(self):
        return iter(self.synthesis_tasks)

    def __getitem__(self, item):
        return self.synthesis_tasks[item]



import functools
import re
from typing import List, Tuple, Union, Dict
from urllib.parse import urlencode

from pydantic import validator

from .base import BasePollyObject
from .params import ContentType
from ..utils.case import to_snake, string_to_snake

__all__ = ['Method']


class Method(BasePollyObject):
    request_method: str
    name: str = None
    endpoint: str = None
    endpoint_template: str = None
    expected_content_types: Union[ContentType, Tuple[ContentType, ...]] = ()
    expected_keys: Union[str, Dict[str, str]] = None
    no_data_on_success: bool = False
    url_params_allowed: bool = False

    @property
    @functools.lru_cache()
    def needed_endpoint_params(self) -> List[str]:
        if self.endpoint_template:
            return re.findall(r'{(.*?)}', self.endpoint_template)
        return []

    def get_url(self, base_url: str, params: dict = None) -> str:
        endpoint_params, url_params = self._filter_params(params)
        endpoint = self._get_endpoint(endpoint_params)

        url = base_url + endpoint

        if url_params:
            url += f'?{urlencode(params)}'
        return url

    def _get_endpoint(self, endpoint_params=None) -> str:
        if self.endpoint:
            return self.endpoint
        elif not self.endpoint_template:
            raise RuntimeError(f'Both self.endpoint and self.endpoint_template can not not be empty')
        elif not self.needed_endpoint_params:
            raise RuntimeError(f'Params to fill not found in endpoint_template: {self.endpoint_template}')
        elif not endpoint_params:
            raise ValueError(f'endpoint_params can not be empty on method {self.name}')

        return self.endpoint_template.format(**endpoint_params)

    def _filter_params(self, params: dict) -> Tuple[dict, dict]:
        params = params or {}
        endpoint_params = {}
        url_params = {}
        for k, v in params.items():
            if v is None:
                raise ValueError(f'Parameter "{k}" is unfilled')

            if k in self.needed_endpoint_params:
                endpoint_params[k] = v
            elif self.url_params_allowed or not self.needed_endpoint_params:
                url_params[k] = v
            else:
                raise ValueError(f'Unexpected parameter: "{k}={v}", allowed params: {self.needed_endpoint_params}')

        unfilled_param = next((param for param in self.needed_endpoint_params if param not in endpoint_params), None)
        if unfilled_param is not None:
            raise ValueError(f'Parameter "{unfilled_param}" not found in "{endpoint_params}",'
                             f' needed params: {self.needed_endpoint_params}')

        return endpoint_params, url_params

    def __set_name__(self, owner, name):
        try:
            self.name = name
        except Exception as e:
            print(e)

    def __str__(self):
        return f'{self.name}: {self.endpoint or self.endpoint_template}'

    def __getattr__(self, item):
        try:
            return BasePollyObject.__getattr__(self, item)
        except AttributeError as e:
            try:
                return self.expected_keys[item]
            except (KeyError, TypeError):
                raise e

    @validator('expected_keys', whole=True)
    def _make_keys_dict(cls, keys) -> dict:
        if isinstance(keys, str):
            return {f'{string_to_snake(keys)}_key': keys}
        elif isinstance(keys, dict):
            return to_snake(keys)
        return {f'{string_to_snake(k)}_key': k for k in keys}

    @validator('expected_content_types', whole=True)
    def _make_tuple_from_content_types(cls, content_types):
        if isinstance(content_types, (ContentType, str)):
            return content_types,
        return content_types



import datetime
from typing import List, Union

from .base import BasePollyObject
from .params import LanguageCode, Alphabet, SpeechMarkTypes
from .speech import SpeechMarksList, Speech

__all__ = ['LexiconAttribute', 'Lexicon', 'LexiconsList']


class LexiconAttribute(BasePollyObject):
    alphabet: Alphabet
    language_code: LanguageCode
    last_modified: datetime.datetime
    lexemes_count: int
    lexicon_arn: str
    size: int


class Lexicon(BasePollyObject):
    name: str = None
    content: str = None
    attributes: LexiconAttribute

    async def synthesize_speech(self, text: str,
                                voice_id: str = None,
                                output_format: str = None,
                                sample_rate: str = None,
                                speech_mark_types: List[Union[SpeechMarkTypes, str]] = None,
                                text_type: str = None,
                                language_code: str = None,
                                auto_convert: bool = None,
                                **convert_params
                                ) -> Union[Speech, SpeechMarksList]:
        """
        Synthesizes speech with self.name in lexicon_names
        """
        return await self.polly.synthesize_speech(text=text,
                                                  voice_id=voice_id,
                                                  output_format=output_format,
                                                  sample_rate=sample_rate,
                                                  speech_mark_types=speech_mark_types,
                                                  text_type=text_type,
                                                  language_code=language_code,
                                                  lexicon_names=[self.name],
                                                  auto_convert=auto_convert,
                                                  **convert_params)


class LexiconsList(BasePollyObject):
    lexicons: List[Lexicon]
    next_token: str = None

    async def synthesize_speech(self, text: str,
                                voice_id: str = None,
                                output_format: str = None,
                                sample_rate: str = None,
                                speech_mark_types: List[Union[SpeechMarkTypes, str]] = None,
                                text_type: str = None,
                                language_code: str = None,
                                auto_convert: bool = None,
                                **convert_params
                                ) -> Union[Speech, SpeechMarksList]:
        """
        Synthesizes speech with self.lexicon_names in lexicon_names
        """

        return await self.polly.synthesize_speech(text=text,
                                                  voice_id=voice_id,
                                                  output_format=output_format,
                                                  sample_rate=sample_rate,
                                                  speech_mark_types=speech_mark_types,
                                                  text_type=text_type,
                                                  language_code=language_code,
                                                  lexicon_names=self.lexicons_names,
                                                  auto_convert=auto_convert,
                                                  **convert_params)

    @property
    def lexicon_names(self) -> list:
        return [lexicon['name'] for lexicon in self.lexicons]

    def __iter__(self):
        return iter(self.lexicons)

    def __getitem__(self, item):
        return self.lexicons[item]



from .case import string_to_camel

DEFAULT_EXCLUDE = {'self', 'cls'}


def generate_params(defaults: dict = None, exclude: set = None, use_camel=True, **kwargs):
    """
    Usage: payload = generate_payload(**locals(), exclude=['foo'], use_camel=False)
    """
    defaults = defaults or {}
    if exclude is None:
        exclude = DEFAULT_EXCLUDE
    else:
        exclude.update(DEFAULT_EXCLUDE)

    params = {}
    for key, value in kwargs.items():
        if key in exclude or key.startswith('_'):
            continue

        if value is None:
            value = defaults.get(key)
            if value is None:
                continue

        elif isinstance(value, dict):
            value = generate_params(**value, exclude=exclude, defaults=defaults)

        elif isinstance(value, list):
            value = [generate_params(**e, exclude=exclude, defaults=defaults)
                     if isinstance(e, dict) else e for e in value]

        params[string_to_camel(key) if use_camel else key] = value

    return params



from .base import BasePollyObject
from .lexicon import Lexicon, LexiconAttribute, LexiconsList
from .method import Method
from .params import (
    AudioFormat, ContentType, LanguageCode, Alphabet, Region,
    TextType, SpeechMarkTypes, SynthesisTaskStatus, VoiceID, Gender
)
from .speech import Speech, SpeechMarks, SpeechMarksList
from .synthesis_task import SynthesisTask, SynthesisTasksList
from .voice import VoicesList, Voice

__all__ = [
    'Alphabet',
    'AudioFormat',
    'BasePollyObject',
    'ContentType',
    'LanguageCode',
    'Lexicon',
    'LexiconAttribute',
    'LexiconsList',
    'Method',
    'Region',
    'Speech',
    'SpeechMarks',
    'SpeechMarkTypes',
    'SpeechMarksList',
    'SynthesisTask',
    'SynthesisTasksList',
    'SynthesisTaskStatus',
    'TextType',
    'Voice',
    'VoiceID',
    'VoicesList'
]



import datetime
import functools
import io
import logging
import os
from typing import List, Union

import aiofiles

from .base import BasePollyObject
from .params import LanguageCode, AudioFormat, ContentType, TextType, SpeechMarkTypes

__all__ = ['Speech', 'SpeechMarks', 'SpeechMarksList']


class ConvertParams(BasePollyObject):
    to_format: str
    out_bitrate: int = None
    duration_in_seconds: int = None
    info: str = None


class Speech(BasePollyObject):
    content_type: ContentType
    request_characters: int
    audio_stream: bytes
    text: str
    voice_id: str
    output_format: AudioFormat
    sample_rate: str = None
    speech_mark_types: SpeechMarkTypes = None
    text_type: TextType = None
    language_code: LanguageCode = None
    lexicon_names: List[str] = None

    converted: bool = False
    converted_stream: bytes = None
    converted_params: ConvertParams = None

    async def convert(self, **kwargs):
        converter = self.polly.converter
        if converter:
            return await self.polly.converter.convert(self, **kwargs)
        raise RuntimeError('Cannot find converter in Polly instance, to use it please specify one')

    @property
    def format(self):
        if self.converted and self.audio_stream == self.converted_stream:
            return self.converted_format
        return self.output_format.split('_')[0].strip()

    @property
    def converted_format(self):
        return self.converted_params.to_format.split('_')[0].strip()

    @property
    @functools.lru_cache()
    def clean_text(self):
        if self.text_type == 'text':
            return self.text

        from ..utils.ssml import clean_text_from_ssml_tags
        return clean_text_from_ssml_tags(self.text)

    @property
    def bytes_io(self):
        return io.BytesIO(self.audio_stream)

    def filename(self, converted=True):
        if converted and self.converted:
            extension = self.converted_format
        else:
            extension = self.format

        name = self.clean_text
        if len(name) > 100:
            name = ' '.join(name[:100].split()[:-1]) + '...'
        name = f'{self.voice_id} - {name}.{extension}'

        return name

    async def save_on_disc(self, filename=None, directory=None, converted=True, overwrite=False):
        if converted and self.converted:
            stream = self.converted_stream
        else:
            stream = self.audio_stream

        if filename is None:
            filename = self.filename(converted=converted)
        if directory is not None:
            filename = '/'.join((directory, filename))

        while not overwrite and os.path.exists(filename):
            name, extension = filename.rsplit('.', maxsplit=1)
            now = datetime.datetime.now().strftime('%S.%f')
            filename = name + f' ({now[:-3]}).' + extension

        logging.debug('Saving %s on disc' % filename)

        async with aiofiles.open(filename, mode='wb') as file:
            await file.write(stream)


class SpeechMarks(BasePollyObject):
    time: int
    type: SpeechMarkTypes
    value: str
    start: int = None
    end: int = None


class SpeechMarksList(BasePollyObject):
    speech_marks: List[SpeechMarks]

    @property
    def vesemes(self):
        return self.get_marks(SpeechMarkTypes.viseme)

    @property
    def sentences(self):
        return self.get_marks(SpeechMarkTypes.sentence)

    @property
    def words(self):
        return self.get_marks(SpeechMarkTypes.word)

    @property
    def ssml(self):
        return self.get_marks(SpeechMarkTypes.ssml)

    def get_marks(self, mark_type: Union[SpeechMarkTypes, str]):
        return [mark for mark in self.speech_marks if mark.type == mark_type]

    def get_marks_values(self, mark_type: Union[SpeechMarkTypes, str]):
        return [mark.value for mark in self.get_marks(mark_type)]

    @property
    def total_time(self):
        return self.speech_marks[-1].time

    def __iter__(self):
        return iter(self.speech_marks)

    def __getitem__(self, item):
        return self.speech_marks[item]



from typing import Optional

from aiopolly.utils.case import to_lower_camel
from ..utils.strenum import StrEnum

__all__ = [
    'Alphabet',
    'AudioFormat',
    'ContentType',
    'Gender',
    'LanguageCode',
    'Region',
    'SpeechMarkTypes',
    'StrEnum',
    'SynthesisTaskStatus',
    'TextType',
    'VoiceID',
]


class TextType(StrEnum):
    text: str
    ssml: str


class SpeechMarkTypes(StrEnum):
    ssml: str
    sentence: str
    viseme: str
    word: str


class SynthesisTaskStatus(StrEnum, converter=to_lower_camel):
    scheduled: str
    in_progress: str
    completed: str
    failed: str


class Alphabet(StrEnum, sep='-'):
    ipa: str
    x_sampa: str
    x_amazon_pinyin: str


class Gender(StrEnum):
    Female: str
    Male: str


class AudioFormat(StrEnum):
    json: str
    mp3: str
    pcm: str
    ogg_vorbis: str


class ContentType(StrEnum, converter=lambda s: s.replace('_', '/', 1), sep='-'):
    application_json: str
    application_octet_stream: str
    application_x_json_stream: str
    audio_mpeg: str
    audio_pcm: str
    audio_ogg: str

    @property
    def audio_format(self) -> Optional[AudioFormat]:
        return AUDIO_CONTENT_TYPE_FORMAT_MAPPING.get(self)


class Region(StrEnum, sep='-'):
    us_east_2: str
    us_east_1: str
    us_west_1: str
    us_west_2: str
    ap_south_1: str
    ap_northeast_2: str
    ap_southeast_1: str
    ap_southeast_2: str
    ap_northeast_1: str
    ca_central_1: str
    cn_northwest_1: str
    eu_central_1: str
    eu_west_1: str
    eu_west_2: str
    eu_west_3: str
    eu_north_1: str
    sa_east_1: str


class LanguageCode(StrEnum, sep='-'):
    arb: str
    cmn_CN: str
    cy_GB: str
    da_DK: str
    de_DE: str
    en_AU: str
    en_GB: str
    en_GB_WLS: str
    en_IN: str
    en_US: str
    es_ES: str
    es_MX: str
    es_US: str
    fr_CA: str
    fr_FR: str
    is_IS: str
    it_IT: str
    ja_JP: str
    hi_IN: str
    ko_KR: str
    nb_NO: str
    nl_NL: str
    pl_PL: str
    pt_BR: str
    pt_PT: str
    ro_RO: str
    ru_RU: str
    sv_SE: str
    tr_TR: str


class VoiceID(StrEnum):
    Aditi: str
    Amy: str
    Astrid: str
    Bianca: str
    Brian: str
    Carla: str
    Carmen: str
    Celine: str
    Chantal: str
    Conchita: str
    Cristiano: str
    Dora: str
    Emma: str
    Enrique: str
    Ewa: str
    Filiz: str
    Geraint: str
    Giorgio: str
    Gwyneth: str
    Hans: str
    Ines: str
    Ivy: str
    Jacek: str
    Jan: str
    Joanna: str
    Joey: str
    Justin: str
    Karl: str
    Kendra: str
    Kimberly: str
    Lea: str
    Liv: str
    Lotte: str
    Lucia: str
    Mads: str
    Maja: str
    Marlene: str
    Mathieu: str
    Matthew: str
    Maxim: str
    Mia: str
    Miguel: str
    Mizuki: str
    Naja: str
    Nicole: str
    Penelope: str
    Raveena: str
    Ricardo: str
    Ruben: str
    Russell: str
    Salli: str
    Seoyeon: str
    Takumi: str
    Tatyana: str
    Vicki: str
    Vitoria: str
    Zeina: str
    Zhiyu: str


AUDIO_CONTENT_TYPES = (
    ContentType.audio_ogg,
    ContentType.audio_pcm,
    ContentType.application_x_json_stream,
    ContentType.audio_mpeg
)

AUDIO_CONTENT_TYPE_FORMAT_MAPPING = {
    ContentType.audio_mpeg: AudioFormat.mp3,
    ContentType.application_x_json_stream: AudioFormat.json,
    ContentType.audio_pcm: AudioFormat.pcm,
    ContentType.audio_ogg: AudioFormat.ogg_vorbis
}



"""
See documentation: https://docs.aws.amazon.com/en_us/polly/latest/dg/supported-ssml.html
"""

import re
from enum import Enum
from typing import Union

from .params import Level, Strength, Volume, Rate, Pitch, Interpretation, DateFormat, Frequency, Duration, SpeechPart
from ...types import LanguageCode, Alphabet

__all__ = [
    'breath',
    'clean_text_from_ssml_tags',
    'drc',
    'emphasis',
    'lang',
    'mark',
    'paragraph',
    'pause',
    'phoneme',
    'prosody',
    'say_as',
    'sentence',
    'soft',
    'ssml_text',
    'sub',
    'timbre',
    'whisper',
]

DEFAULT_EXCLUDE = ('tag', 'text', 'sep')


def make_params(mapping: dict = None,
                exclude: Union[set, list, tuple] = DEFAULT_EXCLUDE,
                allowed: Union[set, list, tuple] = None,
                **params) -> str:
    if isinstance(exclude, set):
        exclude.update(DEFAULT_EXCLUDE)
    elif isinstance(exclude, list):
        exclude.extend(DEFAULT_EXCLUDE)

    if mapping is None:
        mapping = {}

    return ' '.join(
        f'{mapping.get(key, key)}="{value}"'
        for key, value in params.items()
        if value is not None
        and key not in exclude
        and (allowed is None or key in allowed)
    )


def clean_text_from_ssml_tags(text):
    return re.sub('<[^<]+>', '', text)


def ssml_text(*text: str, sep=' '):
    """
    Function witch wraps all positional arguments in <speak></speak> tags

    The <speak> tag is the root element of all Amazon Polly SSML text.
    All SSML-enhanced text must be enclosed within a pair of <speak> tags.

    :param text:
    :param sep:
    :return:
    """
    return f'<speak>{sep.join(text)}</speak>'


def pause(strength: Union[Strength, str] = None,
          seconds: Union[int, float] = None,
          milliseconds: Union[int, float] = None):
    """
    You can set a pause based on strength (equivalent to the pause after a comma, a sentence, or a paragraph),
    or you can set it to a specific length of time in seconds or milliseconds.

    Example:
        code_block python3
            ssml_text(
                f'Mary had a little lamb {pause(seconds=3)}Whose fleece was white as snow.'
            )

    :param strength: affects on break time
    :param seconds: break time in seconds
    :param milliseconds: break time in milliseconds
    """
    tag = '<break {params}/>'

    milliseconds = milliseconds and f'{milliseconds}ms'
    seconds = seconds and f'{seconds}s'

    mapping = {
        'seconds': 'time',
        'milliseconds': 'time'
    }

    return tag.format(params=make_params(**locals()))


def emphasis(*text: str, level: Union[Level, str] = None, sep=' ') -> str:
    """
    To emphasize words, use the <emphasis> tag. Emphasizing words changes the speaking rate and volume.
    More emphasis makes Amazon Polly speak the text louder and slower.
    Less emphasis makes it speak quieter and faster. To specify the degree of emphasis, use the level attribute.

    :param text: text which will be emphasized
    :param level: degree of emphasis
    :param sep: sep between text if more than one given
    """

    tag = '<emphasis {params}>{text}</emphasis>'

    return tag.format(text=sep.join(text), params=make_params(**locals()))


def lang(*text: str, language_code: Union[LanguageCode, str], sep=' ') -> str:
    """
    Specify another language for a specific word, phrase, or sentence with the <lang> tag.
    Foreign language words and phrases are generally spoken better when they are enclosed within a pair of <lang> tags.

    :param text: foreign text
    :param language_code: text language
    :param sep:
    """
    if isinstance(language_code, Enum):
        language_code = language_code.value

    return f'<lang xml:lang="{language_code}">{sep.join(text)}</lang>.'


def mark(name: str) -> str:
    """
    To put a custom tag within the text, use the <mark> tag.
    Amazon Polly takes no action on the tag, but returns the location of the tag in the SSML metadata.

    :param name: name of your mark
    """
    return f'<mark name="{name}"/>'


def paragraph(*paragraphs: str, sep='\n') -> str:
    """
    Provides a longer pause than native speakers usually place at commas or the end of a sentence.

    :param paragraphs: paragraphs
    :param sep:
    :return:
    """

    return sep.join(f'<p>{p}</p>' for p in paragraphs)


def sentence(*sentences: str, sep=' ') -> str:
    """
    Adds a pause between lines or sentences in your text. Using this has the same effect as:
        Ending a sentence with a period (.)
        Specifying a pause with pause(strength='strong')

    :param sentences: sentences
    :param sep:
    :return:
    """
    return sep.join(f'<s>{s}</s>' for s in sentences)


def phoneme(text: str, alphabet: Union[Alphabet, str], ph: str) -> str:
    if isinstance(alphabet, Enum):
        alphabet = alphabet.value
    return f'<phoneme alphabet="{alphabet}" ph="{ph}">{text}</phoneme>'


def prosody(*text: str,
            volume: Union[Volume, str, int] = None,
            rate: Union[Rate, str, int] = None,
            pitch: Union[Pitch, str, int] = None,
            max_duration_s: Union[int, float] = None,
            max_duration_ms: Union[int, float] = None,
            sep=' ') -> str:
    """
    Controls the volume, rate, or pitch of your selected voice.
    Volume, speech rate, and pitch are dependent on the specific voice selected.
    In addition to differences between voices for different languages,
    there are differences between individual voices speaking the same language.
    Because of this, while attributes are similar across all languages,
    there are clear variations from language to language and no absolute value is available.

    :param text: text which will be combined together with :param sep in between
    :param volume: see Volume enum
    :param rate: see Rate enum
    :param pitch: see Pitch enum
    :param max_duration_s: controls how long speech will take when it is synthesized in seconds
    :param max_duration_ms: controls how long speech will take when it is synthesized in milliseconds
    :param sep:
    :return:
    """

    tag = '<prosody {params}>{text}</prosody>'

    if volume and isinstance(volume, int):
        volume = f'{volume}dB'
    if rate and isinstance(rate, int):
        rate = f'{rate}%'
    if pitch and isinstance(pitch, int):
        pitch = f'{pitch}%'

    max_duration_s = max_duration_s and f'{max_duration_s}s'
    max_duration_ms = max_duration_ms and f'{max_duration_ms}ms'

    mapping = {
        'max_duration_s': 'amazon:max-duration',
        'max_duration_ms': 'amazon:max-duration'
    }

    return tag.format(text=sep.join(text), params=make_params(**locals()))


def say_as(*text: str,
           interpret_as: Union[Interpretation, str],
           date_format: Union[DateFormat, str] = None,
           sep=' ') -> str:
    """
    Tells Amazon Polly how to say certain characters, words, and numbers.
    This enables you to provide additional context to eliminate any ambiguity on how Polly should render the text.

    :param text:
    :param interpret_as: see Interpretation enum
    :param date_format: see DateFormat enum
    :param sep:
    :return:
    """
    tag = '<say-as {params}>{text}</say-as>'

    if isinstance(interpret_as, Enum):
        interpret_as = interpret_as.value

    mapping = {
        'date_format': 'format',
        'interpret_as': 'interpret-as'
    }

    return tag.format(text=sep.join(text), params=make_params(**locals()))


def sub(abbreviation: str, alias: str) -> str:
    """
    Substitute a different word (or pronunciation) for selected text such as an acronym or abbreviation.

    :param abbreviation:
    :param alias:
    :return:
    """
    return f'<sub alias="{alias}">{abbreviation}</sub>'


def w(word: str, role: Union[SpeechPart, str]):
    """
    Customizes the pronunciation of words by specifying the word’s part of speech or alternate meaning.

    :param word: word to Customizes
    :param role: see SpeechPart enum
    :return:
    """
    return f'<w role="{role.value if isinstance(role, Enum) else role}">{word}</w>.'


def breath(*text: str,
           duration: Union[Duration, str] = None,
           frequency: Union[Frequency, str] = None,
           volume: Union[Volume, str] = None,
           sep=' ') -> str:
    """
    Adding the Sound of Breathing
    https://docs.aws.amazon.com/en_us/polly/latest/dg/supported-ssml.html#breath-tag

    Natural-sounding speech includes both correctly spoken words and breathing sounds.
    By adding breathing sounds to synthesized speech, you can make it sound more natural.
    The <amazon:breath> and <amazon:auto-breaths> tags provide breaths.

    You have the following options:
        Manual mode: you set the location, length, and volume of a breath sound within the text
        Automated mode: Amazon Polly automatically inserts breathing sounds into the speech output
        Mixed mode: both you and Amazon Polly add breathing sounds

    Examples:

    Single breath:
    .. code_block python3:

        ssml_text(
            f'Sometimes you want to insert only {breath(volume=Volume.x_loud)} a single breath.'
        )

    .. code_block python3

        ssml_text(
            'Sometimes you need',
            breath('to insert one or more average breathes'),
            'so that the text sounds correct.'
        )


    :param text: text parts for automated breath
    :param duration: Controls how loud the breathing sounds. Valid values are listed in Duration enum
    :param frequency: Controls how often breathing sounds occur in the text. Valid values are listed in Frequency enum
    :param volume: Controls how loud the breathing sounds. Valid values are listed in Volume enum
    :param sep: separator for text parts
    """
    if text:
        if duration or frequency or volume:
            raise ValueError('You can\'t any params with auto-breath')
        return f'<amazon:auto-breaths>{sep.join(text)}</amazon:auto-breaths>'
    else:
        tag = '<amazon:breath {params}/>'

        return tag.format(params=make_params(**locals()))


def drc(*text: str, sep=' ') -> str:
    """
    Dynamic Range Compression

    Sets a midrange "loudness" threshold for your audio,
    and increases the volume (the gain) of the sounds around that threshold.
    It applies the greatest gain increase closest to the threshold,
    and the gain increase is lessened farther away from the threshold.

    :param text:
    :param sep:
    :return:
    """
    return f'<amazon:effect name="drc">{sep.join(text)}</amazon:effect>.'


def soft(*text: str, sep=' ') -> str:
    """
    Specifies that input text should be spoken in a softer-than-normal voice.

    :param text:
    :param sep:
    :return:
    """

    return f'<amazon:effect phonation="soft">{sep.join(text)}</amazon:effect>'


def timbre(*text: str, adjust: int = None, absolute: int = None, sep=' ') -> str:
    """
    Timbre is the tonal quality of a voice that helps you tell the difference between voices,
    even when they have the same pitch and loudness. One of the most important physiological
    features that contributes to speech timbre is the length of the vocal tract.
    The vocal tract is a cavity of air that spans from the top of the vocal folds up to the edge of the lips.

    :param text: text to adjust
    :param adjust:
        Adjusts the vocal tract length by a relative percentage change in the current voice.
        For example, +4% or -2%. Valid values range from +100% to -50%. Values outside this range are clipped.
        For example, +111% sounds like +100% and -60% sounds like -50%.
    :param absolute:
        Changes the vocal tract length to an absolute percentage of the tract length of the current voice.
        For example, 110% or 75%. An absolute value of 110% is equivalent to a relative value of +10%.
        An absolute value of 100% is the same as the default value for the current voice.
    :param sep:
    :return:
    """

    if adjust is not None:
        value = f'+{adjust}%' if adjust >= 0 else f'{adjust}%'
    elif absolute is not None:
        value = f'{absolute}%'
    else:
        raise ValueError('Both adjust and absolute cannot be None')

    return f'<amazon:effect vocal-tract-length="{value}">{sep.join(text)}</amazon:effect>'


def whisper(*text: str, sep=' '):
    """
    Indicates that the input text should be spoken in a whispered voice rather than as normal speech.
    This can be used with any of the voices in the Amazon Polly Text-to-Speech portfolio.

    :param text:
    :param sep:
    :return:
    """
    return f'<amazon:effect name=”whispered”>{sep.join(text)}</amazon:effect>'



SERVICE_NAME = 'polly'
BASE_URL_TEMPLATE = 'https://{service_name}.{region}.amazonaws.com'

CHARACTERS_HEADER = 'x-amzn-RequestCharacters'
EXCEPTION_HEADER = 'x-amzn-ErrorType'

TRUST_API_RESPONSES = False
CONVERT_TO_SNAKE_CASE = True



import pathlib
import re

from setuptools import setup, find_packages

try:
    from pip.req import parse_requirements
except ImportError:  # pip >= 10.0.0
    from pip._internal.req import parse_requirements

WORK_DIR = pathlib.Path(__file__).parent
PACKAGE_DIR = WORK_DIR / 'aiopolly'

with open('README.md', 'r') as file:
    README = file.read()


def get_version():
    """
    Read version
    :return: str
    """
    txt = (PACKAGE_DIR / '__init__.py').read_text('utf-8')
    try:
        return re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


def get_requirements():
    requirements_file = WORK_DIR / 'requirements.txt'
    # noinspection PyTypeChecker
    install_reqs = parse_requirements(str(requirements_file), session='hack')
    return [str(ir.requirement) for ir in install_reqs]


setup(
    name='aiopolly',
    version=get_version(),
    author='MrMrRobat',
    author_email='appkiller16@gmail.com',
    description='Asynchronous wrapper for AWS Polly API',
    long_description=README,
    requires_python='>=3.7',
    long_description_content_type='text/markdown',
    url='https://github.com/MrMrRobat/aiopolly',
    packages=find_packages(exclude=['examples']),
    install_requires=get_requirements(),
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
)



import asyncio
from typing import Union, List

from . import api
from . import config
from .. import types
from ..utils import case, json
from ..utils.converter.base import BaseConverter
from ..utils.payload import generate_params


class Methods:
    DeleteLexicon = types.Method(
        endpoint_template='/v1/lexicons/{LexiconName}',
        request_method='DELETE',
        no_data_on_success=True
    )
    DescribeVoices = types.Method(
        endpoint='/v1/voices',
        request_method='GET',
        expected_content_types=types.ContentType.application_json
    )
    GetLexicon = types.Method(
        endpoint_template='/v1/lexicons/{LexiconName}',
        request_method='GET',
        expected_keys='GetLexicon',
        expected_content_types=types.ContentType.application_json
    )
    GetSpeechSynthesisTask = types.Method(
        endpoint_template='/v1/synthesisTasks/{TaskId}',
        request_method='GET',
        expected_keys='SynthesisTask',
        expected_content_types=types.ContentType.application_json
    )
    ListLexicons = types.Method(
        endpoint='/v1/lexicons',
        request_method='GET',
        expected_content_types=types.ContentType.application_json
    )
    ListSpeechSynthesisTasks = types.Method(
        endpoint='/v1/synthesisTasks',
        request_method='GET',
        expected_content_types=types.ContentType.application_json,
    )
    PutLexicon = types.Method(
        endpoint_template='/v1/lexicons/{LexiconName}',
        request_method='PUT',
        no_data_on_success=True
    )
    StartSpeechSynthesisTask = types.Method(
        endpoint='/v1/synthesisTasks',
        request_method='POST',
        expected_keys='SynthesisTask',
        expected_content_types=types.ContentType.application_json
    )
    SynthesizeSpeech = types.Method(
        endpoint='/v1/speech',
        request_method='POST',
        expected_content_types=types.params.AUDIO_CONTENT_TYPES
    )


class Polly(api.AmazonAPIClient):
    """
    You can init this class using one of two methods of authorisation:
        1) Use your access and secret keys with optional params,
        2) Create a .aws/credentials file in your system user Home directory with following data
           and init Polly without auth params:
:
            '''
            [default]
                    aws_access_key_id = your_access_key
                    aws_secret_access_key = your_secret_key
            '''
    """

    methods = Methods
    _requested_characters_header = config.CHARACTERS_HEADER

    def __init__(self,
                 voice_id: str = None,
                 output_format: str = None,
                 sample_rate: str = None,
                 speech_mark_type: str = None,
                 text_type: Union[List[str], str] = None,
                 language_code: str = None,
                 lexicon_names: List[str] = None,
                 engine: str = None,
                 speech_mark_types: List[str] = None,
                 output_s3_key_prefix: str = None,
                 output_s3_bucket_name: str = None,
                 sns_topic_arn: str = None,
                 include_additional_language_codes: bool = False,
                 region: Union[types.Region, str] = types.Region.eu_central_1.value,
                 access_key: str = None,
                 secret_key: str = None,
                 converter: BaseConverter = None,
                 loop: asyncio.AbstractEventLoop = None,
                 **defaults):
        """
        Auth and AWS params:
            :param credentials: instance of botocore.credentials.Credentials class;
            :param session: instance of botocore.session.Session with credentials
            :param access_key: AWS access key, requires :param secret_key;
            :param secret_key: AWS secret key, requires :param access_key;
            :param region: AWS server region, default is 'eu-central-1'. For available regions see:
                https://docs.aws.amazon.com/AmazonRDS/latest/UserGuide/Concepts.RegionsAndAvailabilityZones.html

        Default API params, used when method params remain empty:
            For original docs see: https://docs.aws.amazon.com/en_us/polly/latest/dg/API_Operations.html

            :param voice_id: speech voice_id;
            :param output_format: the format in which the returned output will be encoded;
            :param sample_rate: audio frequency specified in Hz;
            :param lexicon_names: list of one or more pronunciation lexicon names;
            :param engine: Speech engine to use, either 'standard' or 'neural'. Some voices are not available with 'neural'.
            :param text_type: specifies whether the input text is plain text or SSML;
            :param language_code: optional language code for the Synthesize Speech request.
            :param output_s3_key_prefix: the Amazon S3 key prefix for the output speech file;
            :param output_s3_bucket_name: amazon S3 bucket name to which the output file will be saved.
            :param sns_topic_arn: ARN for the SNS topic for providing status notification for a speech synthesis task.
            :param include_additional_language_codes: value indicating whether to return any bilingual speech that use
                the specified language as an additional language.
        """

        super().__init__(
            region=region,
            access_key=access_key,
            secret_key=secret_key,
            loop=loop
        )

        self.converter = converter

        # Setting default params
        self.defaults = dict(
            voice_id=voice_id,
            output_format=output_format,
            sample_rate=sample_rate,
            speech_mark_type=speech_mark_type,
            text_type=text_type,
            language_code=language_code,
            lexicon_names=lexicon_names,
            engine=engine,
            speech_mark_types=speech_mark_types,
            output_s3_key_prefix=output_s3_key_prefix,
            output_s3_bucket_name=output_s3_bucket_name,
            sns_topic_arn=sns_topic_arn,
            include_additional_language_codes=include_additional_language_codes
        )
        self.defaults.update(defaults)
        self.defaults = {key: value for key, value in self.defaults.items() if value is not None}

    async def delete_lexicon(self, lexicon_name: str) -> None:
        """
        Deletes the specified pronunciation lexicon stored in an AWS Region.
        A lexicon which has been deleted is not available for speech synthesis,
        nor is it possible to retrieve it using either the GetLexicon or ListLexicon APIs.

        See: https://docs.aws.amazon.com/en_us/polly/latest/dg/API_DeleteLexicon.html

        :param lexicon_name: The name of the lexicon to delete. Must be an existing lexicon in the region.
        """
        params = generate_params(lexicon_name=lexicon_name)

        await self.request(self.methods.DeleteLexicon, params=params)

    async def describe_voices(self, include_additional_language_codes: bool = None,
                              language_code: str = None, next_token: str = None) -> types.VoicesList:
        """
        Returns the list of speech that are available for use when requesting speech synthesis.
        Each voice speaks a specified language, is either male or female, and is identified by an ID,
        which is the ASCII version of the voice name.

        See: https://docs.aws.amazon.com/en_us/polly/latest/dg/API_DescribeVoices.html

        :param include_additional_language_codes: Boolean value indicating whether to return any bilingual speech
               that use the specified language as an additional language.
        :param language_code: The language identification tag for filtering the list of speech returned.
        :param next_token: An opaque pagination token returned from the previous DescribeVoices operation.
        """
        params = generate_params(**locals(), defaults=self.defaults)
        result, response = await self.request(self.methods.DescribeVoices, params=params)

        return types.VoicesList(**result)

    async def get_lexicon(self, lexicon_name: str) -> types.Lexicon:
        """
        Returns the content of the specified pronunciation lexicon stored in an AWS Region.

        See: https://docs.aws.amazon.com/en_us/polly/latest/dg/API_GetLexicon.html

        :param lexicon_name: Name of the lexicon.
        """

        method = self.methods.GetLexicon
        params = generate_params(lexicon_name=lexicon_name)
        result, response = await self.request(method, params=params)
        lexicon_attributes = result[method.lexicon_attributes_key]

        return types.Lexicon(**result[method.lexicon_key], attributes=lexicon_attributes)

    async def get_speech_synthesis_task(self, task_id: str) -> types.SynthesisTask:
        """
        Retrieves a specific SpeechSynthesisTask object based on its TaskID.
        This object contains information about the given speech synthesis task,
        including the status of the task, and a link to the S3 bucket containing the output of the task.

        See: https://docs.aws.amazon.com/en_us/polly/latest/dg/API_GetSpeechSynthesisTask.html

        :param task_id: The Amazon Polly generated identifier for a speech synthesis task.
        """

        method = self.methods.GetSpeechSynthesisTask
        params = generate_params(task_id=task_id)
        result, response = await self.request(self.methods.GetSpeechSynthesisTask, params=params)

        return types.SynthesisTask(**result[method.synthesis_task_key])

    async def list_lexicons(self, next_token: str = None) -> types.LexiconsList:
        """
        Returns a list of pronunciation lexicons stored in an AWS Region.

        See: https://docs.aws.amazon.com/en_us/polly/latest/dg/API_ListLexicons.html

        :param next_token: An opaque pagination token returned from previous ListLexicons operation.
               If present, indicates where to continue the list of lexicons.
        """

        params = generate_params(next_token=next_token)
        result, response = await self.request(self.methods.ListLexicons, params=params)

        return types.LexiconsList(**result)

    async def list_speech_synthesis_tasks(self, max_results: int = None,
                                          next_token: str = None,
                                          status: str = None) -> types.SynthesisTasksList:
        """
        Returns a list of SpeechSynthesisTask objects ordered by their creation date.
        This operation can filter the tasks by their status, for example,
        allowing users to list only tasks that are completed.

        See: https://docs.aws.amazon.com/en_us/polly/latest/dg/API_ListSpeechSynthesisTasks.html

        :param max_results: Maximum number of speech synthesis tasks returned in a List operation. (1-100)
        :param next_token: The pagination token to use in the next request to continue the listing
        :param status: Status of the speech synthesis tasks returned in a List operation
        """

        params = generate_params(**locals())
        result, response = await self.request(self.methods.ListSpeechSynthesisTasks, params=params)

        return types.SynthesisTasksList(**result)

    async def put_lexicon(self, lexicon_name: str, content: str):
        """
        Stores a pronunciation lexicon in an AWS Region.
        If a lexicon with the same name already exists in the region, it is overwritten by the new lexicon.
        Lexicon operations have eventual consistency, therefore, it might take some time
        before the lexicon is available to the SynthesizeSpeech operation.

        :param lexicon_name: Name of the lexicon. The name must follow the regular express format [0-9A-Za-z]{1,20}.
               That is, the name is a case-sensitive alphanumeric string up to 20 characters long.
        :param content: Content of the PLS lexicon as string data.
        """

        params = generate_params(lexicon_name=lexicon_name)
        payload = generate_params(content=content)

        await self.request(self.methods.PutLexicon, payload=payload, params=params)

    async def start_speech_synthesis_task(self, text: str,
                                          output_s3_key_prefix: str = None,
                                          output_s3_bucket_name: str = None,
                                          voice_id: str = None,
                                          output_format: str = None,
                                          sample_rate: str = None,
                                          speech_mark_types: str = None,
                                          text_type: str = None,
                                          language_code: str = None,
                                          lexicon_names: str = None,
                                          engine: str = None
                                          ) -> types.SynthesisTask:
        """
        Allows the creation of an asynchronous synthesis task, by starting a new SpeechSynthesisTask.
        This operation requires all the standard information needed for speech synthesis,
        plus the name of an Amazon S3 bucket for the service to store the output of the synthesis task
        and two optional parameters (OutputS3KeyPrefix and SnsTopicArn). Once the synthesis task is created,
        this operation will return a SpeechSynthesisTask object, which will include an identifier of this task
        as well as the current status.

        :param text: The input text to synthesize.
        :param output_s3_key_prefix: The Amazon S3 key prefix for the output speech file.
        :param output_s3_bucket_name: Amazon S3 bucket name to which the output file will be saved.
        :param voice_id: Voice ID to use for the synthesis.
        :param output_format: The format in which the returned output will be encoded.
        :param sample_rate: The audio frequency specified in Hz.
        :param speech_mark_types: The type of speech marks returned for the input text.
        :param text_type: Specifies whether the input text is plain text or SSML. The default value is plain text.
        :param language_code: Optional language code for the Synthesize Speech request.
        :param lexicon_names: List of one or more pronunciation lexicon names to apply during synthesis
        :param engine: Speech engine to use, either 'standard' or 'neural'. Some voices are not available with 'neural'.
        :return:
        """

        payload = generate_params(**locals(), defaults=self.defaults)
        result, response = await self.request(self.methods.StartSpeechSynthesisTask, payload=payload)
        return types.SynthesisTask(**result)

    async def synthesize_speech(self, text: str,
                                voice_id: str = None,
                                output_format: Union[types.AudioFormat, str] = None,
                                sample_rate: str = None,
                                speech_mark_types: List[Union[types.SpeechMarkTypes, str]] = None,
                                text_type: Union[types.TextType, str] = None,
                                language_code: Union[types.LanguageCode, str] = None,
                                lexicon_names: list = None,
                                auto_convert: bool = None,
                                engine: str = None,
                                **converter_params
                                ) -> Union[types.Speech, types.SpeechMarksList]:
        """
        Synthesizes UTF-8 input, plain text or SSML, to a stream of bytes. SSML input must be valid, well-formed SSML.
        Some alphabets might not be available with all the speech
        (for example, Cyrillic might not be read at all by English speech) unless phoneme mapping is used.

        :param text: Input text to synthesize.
        :param voice_id: Voice ID to use for the synthesis.
        :param output_format: The format in which the returned output will be encoded.
        :param sample_rate: The audio frequency specified in Hz.
        :param speech_mark_types: The type of speech marks returned for the input text.
        :param text_type: Specifies whether the input text is plain text or SSML. The default value is plain text.
        :param language_code: Optional language code for the Synthesize Speech request.
        :param lexicon_names: List of one or more pronunciation lexicon names to apply during synthesis
        :param engine: Speech engine to use, either 'standard' or 'neural'. Some voices are not available with 'neural'.

        :param auto_convert: param indicate whether speech will be auto converted after synthesis
        :param converter_params: params which will be placed in self.converter.convert method
        """
        payload = generate_params(**locals(), defaults=self.defaults, use_camel=False,
                                  exclude={'auto_convert', 'converter_params'})

        content, response = await self.request(self.methods.SynthesizeSpeech, payload=case.to_camel(payload))

        if response.content_type == types.ContentType.application_x_json_stream:
            return types.SpeechMarksList(
                speech_marks=[json.loads(line) for line in content.split(b'\n')[:-1]]
            )

        speech = types.Speech(
            content_type=response.content_type,
            request_characters=response.headers[self._requested_characters_header],
            audio_stream=content,
            **payload
        )

        if self.converter and (auto_convert or self.converter.auto_convert and auto_convert is not False):
            return await self.converter.convert(speech, **converter_params)
        elif auto_convert or converter_params:
            raise RuntimeError(f'Cannot find converter in {self}, to use it please specify one')

        return speech



import abc

from ...types import Speech

__all__ = ['BaseConverter']


class BaseConverter(abc.ABC):
    auto_convert: bool
    keep_original: bool

    @abc.abstractmethod
    async def convert(self, speech: Speech, **kwargs) -> Speech:
        """
        Example:
            async def convert(self, speech: Speech, to_format=None, out_bitrate=None) -> Speech:
                converted_audio, duration, info = self._convert_audio(speech.audio_stream, to_format, out_bitrate)

                if not self.keep_original:
                    speech.audio_stream = converted_audio
                speech.converted_stream = converted_audio

                speech.converted = True
                speech.converted_params = {
                    'to_format': to_format or self.to_format
                    'out_bitrate': out_bitrate or self.bitrate
                    'duration_in_seconds': duration
                    'info': info
                }
                return speech

        :param speech: speech which needs to be converted
        :param kwargs: any convert params
        :return: Speech with additional params and converted audio
        """
        pass



import datetime
import functools
import logging

from pydantic import BaseModel, Extra
from pydantic.json import timedelta_isoformat

from ..utils import json
from ..utils.case import to_camel

__all__ = ['BasePollyObject']

log = logging.getLogger('aiopolly')


class BasePollyObject(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow
        validate_assignment = True
        allow_population_by_alias = True
        alias_generator = to_camel

        json_encoders = {
            datetime.datetime: lambda v: (v.replace(tzinfo=None) - datetime.datetime(1970, 1, 1)).total_seconds(),
            datetime.timedelta: timedelta_isoformat,
        }

    @property
    @functools.lru_cache()
    def polly(self):
        """
        :rtype: aiopolly.polly.Polly
        """
        from .. import Polly
        polly = Polly.get_current()
        if polly is None:
            raise RuntimeError("Can't get polly instance from context. "
                               "You can fix it with setting current instance: "
                               "'Polly.set_current(polly_instance)'")
        return polly

    def raw_dict(self):
        """
        :return: raw data that we get from AWS Polly API
        """
        return json.loads(self.dict(by_alias=True))

    def __hash__(self):
        def _hash(obj):
            buf = 0
            if isinstance(obj, list):
                for item in obj:
                    buf += _hash(item)
            elif isinstance(obj, dict):
                for dict_key, dict_value in obj.items():
                    buf += hash(dict_key) + _hash(dict_value)
            else:
                try:
                    buf += hash(obj)
                except TypeError:  # Skip unhashable objects
                    pass
            return buf

        result = 0
        for key, value in sorted(self.fields.items()):
            result += hash(key) + _hash(value)

        return result



import functools
import re
from typing import Union, Callable, Any


def string_to_camel(string: str):
    return ''.join(x[:1].upper() + x[1:] for x in string.split('_'))


def string_to_lower_camel(string: str):
    in_camel = string_to_camel(string)
    if len(in_camel) > 1:
        return in_camel[0].lower() + in_camel[1:]
    elif string:
        return in_camel.lower()


def string_to_snake(string: str):
    result = re.sub('([A-Z]+)', r'_\1', string).lower()
    if result.startswith('_'):
        return result[1:]
    return result


def to_case(
        obj: Union[str, dict, list, Any],
        ignore_string=False,
        any_type=False,
        str_converter: Callable = string_to_snake
) -> Union[str, dict, list, Any]:
    if isinstance(obj, str):
        return obj if ignore_string else str_converter(obj)
    elif isinstance(obj, dict):
        return {
            str_converter(key):
            to_case(value, ignore_string=True, any_type=True, str_converter=str_converter)
            for key, value in obj.items()
        }
    elif isinstance(obj, list):
        return [
            to_case(element, ignore_string=True, any_type=True, str_converter=str_converter)
            for element in obj
        ]
    elif any_type:
        return obj

    raise ValueError(f'Unexpected type {type(obj)}. Supported types: str, dict, list')


to_snake = functools.partial(to_case, str_converter=string_to_snake)
to_camel = functools.partial(to_case, str_converter=string_to_camel)
to_lower_camel = functools.partial(to_case, str_converter=string_to_lower_camel)



"""Using json from aiogram see: https://github.com/aiogram/aiogram/blob/dev-2.x/aiogram/utils/json.py"""

import importlib
import os

JSON = 'json'
RAPIDJSON = 'rapidjson'
UJSON = 'ujson'

# Detect mode
mode = JSON
for json_lib in (RAPIDJSON, UJSON):
    if 'DISABLE_' + json_lib.upper() in os.environ:
        continue

    try:
        json = importlib.import_module(json_lib)
    except ImportError:
        continue
    else:
        mode = json_lib
        break
if mode == RAPIDJSON:
    def dumps(data, ensure_ascii=False, number_mode=json.NM_NATIVE,
              datetime_mode=json.DM_ISO8601 | json.DM_NAIVE_IS_UTC, **kwargs):
        return json.dumps(data, ensure_ascii=ensure_ascii, number_mode=number_mode,
                          datetime_mode=datetime_mode, **kwargs)


    def loads(data):
        return json.loads(data, number_mode=json.NM_NATIVE,
                          datetime_mode=json.DM_ISO8601 | json.DM_NAIVE_IS_UTC)

elif mode == UJSON:
    def loads(data):
        return json.loads(data)


    def dumps(data, ensure_ascii=False, **kwargs):
        return json.dumps(data, ensure_ascii=ensure_ascii, **kwargs)

else:
    import json


    def dumps(data, ensure_ascii=False, **kwargs):
        return json.dumps(data, ensure_ascii=ensure_ascii, **kwargs)


    def loads(data, **kwargs):
        return json.loads(data, **kwargs)
